#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

set -e

#GLOBALS
SCRIPT_DIR=$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)
if [[ -z "${SCRIPT_DIR}" ]]; then
  error "Could not determine script path"
fi
MODULES_DIR=$(cd "${SCRIPT_DIR}/../../Modules" && pwd)
if [[ -z "${MODULES_DIR}" ]]; then
  error "Could not determine modules path"
fi

BASE_NAME=$(basename "$0")
COMMAND_NAME="${COMMAND_NAME:-$BASE_NAME}"

readonly SCRIPT_DIR
readonly MODULES_DIR

export KUBECONFIG=${KUBECONFIG:-/etc/rancher/rke2/rke2.yaml}
export DOT_FOLDER=~/.uipath
[[ -d "${DOT_FOLDER}" ]] || mkdir -p "${DOT_FOLDER}"

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/utils.sh

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/argocd_utils.sh

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/cluster_utils.sh

ARGOCD_CONFIG_FILE="${MODULES_DIR}"/argocd_config.json
DEFAULT_DB_NAMES="$(cat "${MODULES_DIR}"/default_db_names.json)"

function display_usage() {
  echo
  echo "***************************************************************************************"
  echo
  echo "Get or generate cluster_config.json file"
  echo 
  echo "Usage:"
  echo "  $COMMAND_NAME [flags]"
  echo 
  echo "Flags:"
  echo "  -i|--input                   [Optional] Path to the old cluster_config.json file."
  echo "                                  If not provided then, you may have to manually add few paramaters"                     
  echo "  -o|--outpath                 [Optional] Path when the file should be stored."
  echo "                                  If not provided then file will be store in the current directory"
  echo "  -h|--help                    Display help" 
  echo
  echo "***************************************************************************************"
  echo
}

function parse_long_args() {
  while (("$#")); do
    case "$1" in
    -d | --debug)
      info "Running in debug"
      shift
      set -x
      ;;

    -h | --help)
      display_usage
      shift
      exit 0
      ;;
    
    -o|--outpath)
      if [ -n "$2" ]; then
        OUTPATH="$2"
        shift 2
      else
        display_usage
        error "Argument for $1 is missing"
        shift 2
      fi
      ;;
    
    -i|--input)
      if [ -n "$2" ]; then
        INFILE="$2"
        shift 2
      else
        display_usage
        error "Argument for $1 is missing"
        shift 2
      fi
      ;;

    --kubeconfig)
      shift 2
      ;;

    -*|*) # unsupported flags
      display_usage
      error "Error: Unsupported flag $1"
      exit 1
    ;;
    esac
  done

  if [[ -z ${OUTPATH} ]]; then 
    OUTPATH=$(mktemp -p "${DOT_FOLDER}" -t cluster_config.XXXXXXXXXX.json)
    info "Outpath is not provided, config file will be generated in: $OUTPATH"
  fi
  if [[ ! -f ${INFILE} ]]; then 
    warn "Old or existing cluster_config.json file is not provided, you may have to edit the file manually with the missing values"
  else 
    jq '.' "${INFILE}" &> /dev/null || error "Input file ${INFILE} is not a valid json and cannot be parsed"
  fi

  echo
}

function get_offline_registry_credentials() {
  local config=$1

  username=$(get_cluster_config "infra.docker_registry.username")
  password=$(get_cluster_config "infra.docker_registry.password")

  # also check the old location used to store these
  [[ -n "${username}" ]] \
    || username=$(kubectl -n uipath get secrets docker-registry-credentials -o jsonpath='{.data.username}' --ignore-not-found | base64 -d -w 0) || true
  [[ -n "${password}" ]] \
    || password=$(kubectl -n uipath get secrets docker-registry-credentials -o jsonpath='{.data.password}' --ignore-not-found | base64 -d -w 0) || true

  config=$(jq ".infra.docker_registry.username=\"$username\""  <<< "$config")
  config=$(jq ".infra.docker_registry.password=\"$password\""  <<< "$config")

  echo "$config"
}

function get_remote_registry_credentials() {
  local config=$1
  docker_username=$(get_cluster_config "registries.docker.username")
  docker_password=$(get_cluster_config "registries.docker.password")
  docker_pullsecret=$(get_cluster_config "registries.docker.pull_secret_value")
  
  helm_username=$(get_cluster_config "registries.helm.username")
  helm_password=$(get_cluster_config "registries.helm.password")

  # also check the old location used to store these
  [[ -n "${docker_username}" ]] \
    || docker_username=$(kubectl -n argocd get secrets helm-credentials -o jsonpath='{.data.username}' --ignore-not-found | base64 -d -w 0) || true
  [[ -n "${docker_password}" ]] \
    || docker_password=$(kubectl -n argocd get secrets helm-credentials -o jsonpath='{.data.password}' --ignore-not-found | base64 -d -w 0) || true
  [[ -n "${docker_pullsecret}" ]] \
    || docker_pullsecret=$(argocd app get fabric-installer --output json | jq -r '.spec.source.helm.parameters[] | select(.name=="global.dockerRegistry.imagePullSecretValue").value') || true
  [[ -n "${helm_username}" ]] \
    || helm_username=$(kubectl -n argocd get secrets helm-credentials -o jsonpath='{.data.username}' --ignore-not-found | base64 -d -w 0) || true
  [[ -n "${helm_password}" ]] \
    || helm_password=$(kubectl -n argocd get secrets helm-credentials -o jsonpath='{.data.password}' --ignore-not-found | base64 -d -w 0) || true

  [[ -n "${docker_username}" ]] && { config=$(jq ".registries.docker.username=\"$docker_username"\"  <<< "$config") || return 1; } 
  [[ -n "${docker_password}" ]] && { config=$(jq ".registries.docker.password=\"$docker_password"\"  <<< "$config") || return 1; }
  [[ -n "${docker_pullsecret}" ]] && { config=$(jq ".registries.docker.pull_secret_value=\"$docker_pullsecret"\"  <<< "$config") || return 1; } 

  [[ -n "${helm_username}" ]] && { config=$(jq ".registries.helm.username=\"$helm_username"\"  <<< "$config") || return 1; }
  [[ -n "${helm_username}" ]] && { config=$(jq ".registries.helm.password=\"$helm_password"\"  <<< "$config") || return 1; }
  echo "$config"

}

function get_fixed_rke_address() {
  local config=$1

  fixed_rke_address=$(get_cluster_config "fixed_rke_address")
  config=$(jq ".fixed_rke_address=\"$fixed_rke_address\""  <<< "$config")

  echo "$config"
}

function get_rke_token() {
  local config=$1

  rke_token=$(get_cluster_config "rke_token")
  config=$(jq ".rke_token=\"$rke_token\""  <<< "$config")
  echo "$config"
}

function update_proxy_field() {
  local config=$1
  local no_proxy fqdn effective_no_proxy

  no_proxy=$(jq -r '.proxy.no_proxy // ""' <<< "$config")
  fqdn=$(jq -r .fqdn <<< "$config")\

  # Check the old place used to store the user list for no_proxy, for backwards compatibility
  effective_no_proxy=$(argocd app get uipath --output json | jq '.spec.source.helm.parameters[] | select(.name=="global.proxy.noproxy").value')
  if [[ -z "${no_proxy}" ]] && [[ "${effective_no_proxy}" != *"PLACEHOLDER"* ]]; then
    # Keep only the user input (stored before the fdqn)
    no_proxy="${effective_no_proxy%%\\,"$fqdn"*}"
    # Remove the comma escapes added by ArgoCD
    no_proxy="${no_proxy//\\/}"
  fi
  
  echo "$no_proxy"
  [[ -z "${no_proxy}" ]] || config=$(jq ".proxy.no_proxy=\"$no_proxy\""  <<< "$config")

  echo "$config"
}

function build_sql_templates() {
  local template_type template_regex templates most_frequent_template json_paths json_paths_arr get_cmd delete_cmd config set_template

  set_template=0
  template_type="$1"  
  config="$2"

  case "${template_type}" in
  net)  template_regex=".*Server=.*Initial Catalog.*$"
        template_json_key=".sql_connection_string_template"
        template_db_sed="s/Initial Catalog=[^;]*;/Initial Catalog=DB_NAME_PLACEHOLDER;/gi"
    ;;
  jdbc) template_regex=".*jdbc:.*database=.*$"
        template_json_key=".sql_connection_string_template_jdbc"
        template_db_sed="s/database=[^;]*;/database=DB_NAME_PLACEHOLDER;/gi"
    ;;
  odbc) template_regex=".*SERVER=.*DATABASE=.*$"
        template_json_key=".sql_connection_string_template_odbc"
        template_db_sed="s/DATABASE=[^;]*;/DATABASE=DB_NAME_PLACEHOLDER;/gi"
    ;;
  *) error "Bad sql template template type: ${template_type}"
    ;;
  esac

  # transform every connection string into a possible template
  templates=$(echo "${config}" | jq '..|.sql_connection_str?|select(length>0)' \
            | grep -Ei "${template_regex}" \
            | sed -r "${template_db_sed}")
  templates_count=$(echo "${templates}" | wc -l)
  # check which one is used the most
  most_frequent_template=$(echo "${templates}" | sort | uniq -c | sort -rn | head -n 1 | sed -E 's/^ *[0-9]+ //g')
  most_frequent_template_count=$(echo "${templates}" | grep -F -c "${most_frequent_template}")

  # do nothing if there are more than 1 possible templates and nothing repeats
  if (( ! most_frequent_template_count > 1 )) && (( templates_count > 1 )); then
    echo "${config}"
    return 0
  fi

  # iterate through all 'sql_connection_str' properties and remove them if they match the template
  json_paths=$(echo "${config}" | jq -r 'path(..)|[.[]|tostring]|join(".")' | grep "sql_connection_str")
  mapfile -t json_paths_arr <<< "${json_paths}"

  for path in "${json_paths_arr[@]}"
  do
    get_cmd="echo \$config | jq '.${path}'"
    value=$(eval "${get_cmd}")
    default_db_name="$(echo "${DEFAULT_DB_NAMES}" | jq -r ".[] | select(.config_path == \".${path}\").db_name")"
    rendered_template="${most_frequent_template//DB_NAME_PLACEHOLDER/$default_db_name}"

    delete_cmd="echo \$config  | jq 'del(.${path})'"    
    [[ "${value}" == "${rendered_template}" ]] && { config="$(eval "${delete_cmd}")"; set_template=1; }
  done

  (( set_template == 1 )) && config=$(jq "${template_json_key}=${most_frequent_template}"  <<< "$config")

  # remove any {} elemets that might result after removing redundant 'sql_connection_str' keys
  config=$(echo "${config}" | jq 'del(.. | select(. == {}))')

  echo "${config}"
}

function get_cluster_config_json_file() {
  argocd_app="uipath"
  argocd_config_map=$(cat "${ARGOCD_CONFIG_FILE}")
  app_params=$(argocd app get $argocd_app --output json | jq .spec.source.helm.parameters)

  config="{}"
  while IFS=$'\t' read -r key value;
  do
    param=$(echo "$argocd_config_map" | jq -r ".[] | select(.argocd_param == \"$key\").config_param")
    is_encoded="$(echo "$argocd_config_map" | jq ".[] | select(.argocd_param == \"$key\").is_encoded")"
    if [[ $is_encoded == '"true"' ]]; then
      value=$(base64 -d <<< "$value")
    fi

    # Ignore placeholder values (b64 too)
    [[ "$value" == "PLACEHOLDER" || "$value" == "UExBQ0VIT0xERVI=" ]] && continue
    
    # Don't quote booleans or numbers
    if [[ -n "$param" && -n "$value" ]]; then
      if [[ "$value" == "false" ]] || [[ "$value" == "true" ]] || [[ "$value" =~ ^[0-9]+$ ]]; then
        config=$(jq ".$param=$value"  <<< "$config")
        continue
      fi
      config=$(jq ".$param=\"$value\""  <<< "$config")
    fi
  done <<< "$(echo "$app_params" | jq -r '.[] | [.name, .value] | @tsv')"

  #shellcheck disable=SC2015
  config=$(get_offline_registry_credentials "$config") &&
  config=$(get_remote_registry_credentials "$config") &&
  config=$(get_fixed_rke_address "$config") &&
  config=$(get_rke_token "$config") &&
  config=$(update_proxy_field "$config") &&
  config=$(build_sql_templates "net" "$config") &&
  config=$(build_sql_templates "jdbc" "$config") &&
  config=$(build_sql_templates "odbc" "$config") || error_with_code "12015"

  echo "$config" > "$OUTPATH"

  #shellcheck disable=SC2005
  [[ -f $INFILE ]] && echo "$(jq -s '.[0] * .[1]' "${INFILE}" "${OUTPATH}")" > "${OUTPATH}"

  info "Cluster configuration is stored in file: $OUTPATH"
}

function main() {
  parse_long_args "$@"
  get_cluster_config_json_file
}

main "$@"
