#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

## Variable via environment variable
LOG_RETENTION_DAYS=${LOG_RETENTION_DAYS:-30}
DIFF_RETENTION_DAYS=${DIFF_RETENTION_DAYS:-7}
NFS_PATH="${NFS_PATH:-/nfs}"
EXCLUDE_BUCKETS="${EXCLUDE_BUCKETS:-}" ## comma separated list
INCLUDE_BUCKETS="${INCLUDE_BUCKETS:-}" ## comma separated list
EXCLUDE_BUCKETS_PATTERN="${EXCLUDE_BUCKETS_PATTERN:-}"  ## grep perl regex
INCLUDE_BUCKETS_PATTERN="${INCLUDE_BUCKETS_PATTERN:-}"  ## grep perl regex
RESTORE_EPOCH="${RESTORE_EPOCH:-latest}"
RESTORE_TEMP_PATH="${RESTORE_TEMP_PATH:-${NFS_PATH}/tmp}"
OP="${OP:-backup}"

set -e

function clean_up_old_dir(){
  local epoch_threshold="$1"
  local dst_dir="$2"
  local dir

  [[ ! -d $dst_dir ]] && return
  for dir in "$dst_dir"/*/; do
    dir=$(basename "$dir")
    if [[ -d ${dst_dir}/$dir && $dir -lt $epoch_threshold ]] ; then
      echo "cleaning up ${dst_dir}/$dir"
      \rm -rf "${dst_dir:?}/${dir:?}"
    fi
  done
}

function call_s3_api() {
  local command="$1"
  local url="$2"
  local file="${3:--}"
  local args=()

  if [ "${url:0:5}" != "s3://" ]; then
    echo "Need an s3 url"
    return 1
  fi
  local path="${url:4}"

  if [ -z "${AWS_ACCESS_KEY-}"  ]; then
    echo "Need AWS_ACCESS_KEY to be set"
    return 1
  fi

  if [ -z "${AWS_SECRET_KEY-}" ]; then
    echo "Need AWS_SECRET_KEY to be set"
    return 1
  fi

  local method md5 args
  case "$command" in
  get)
    method="GET"
    md5=""
    args+=(-o "$file")
    ;;
  put)
    method="PUT"
    if [ ! -f "$file" ]; then
      echo "file not found"
      exit 1
    fi
    md5="$(openssl md5 -binary "$file" | openssl base64)"
    args+=(-T "$file") 
    args+=(-H Content-MD5:"$md5")
    ;;
  *)
    echo "Unsupported command"
    return 1
  esac

  local date
  local string_to_sign
  local signature

  date="$(date -u '+%a, %e %b %Y %H:%M:%S +0000')"
  printf -v string_to_sign "%s\n%s\n\n%s\n%s" "$method" "$md5" "$date" "$path"
  signature=$(echo -n "$string_to_sign" | openssl sha1 -binary -hmac "${AWS_SECRET_KEY}" | openssl base64)
  local authorization="AWS ${AWS_ACCESS_KEY}:${signature}"

  curl "${args[@]}" -s -f -H Date:"${date}" -H Authorization:"${authorization}" http://rook-ceph-rgw-rook-ceph.rook-ceph.svc"${path}"
}


function get_bucket_config() {
  local bucket_name="$1"
  local bucket_config="$2"
  bucket_config=$(call_s3_api "get" "s3://${bucket_name}/?${bucket_config}")
  [[ -z $bucket_config ]] || echo "$bucket_config"
}

function put_bucket_config() {
  local bucket_name="$1"
  local bucket_config="$2"
  local data_file="$3"
  call_s3_api "put" "s3://${bucket_name}/?${bucket_config}" "$data_file"
}

function restore_bucket_config() {
  local bucket_name="$1"
  local bucket_config="$2"
  local data_file="$3"
  local return_code=0

  echo "Restoring bucket: '${bucket_name}' ${bucket_config^^}"
  if put_bucket_config "${bucket_name}" "${bucket_config}" "${data_file}"; then
    echo "Successfully restored bucket: '${bucket}' ${bucket_config^^}"
  else
    echo "Failure while restoring bucket: '${bucket}' ${bucket_config^^}"
    return_code=1
  fi
  return "$return_code"
}

function get_bucket_owner(){
  local bucket_name="$1"
  local bucket_owner
  bucket_owner=$(kubectl --cache-dir /tmp -n rook-ceph exec deploy/rook-ceph-tools -- radosgw-admin bucket stats --bucket "${bucket}" | jq '.owner' -r)
  [[ -z $bucket_owner ]] || echo "$bucket_owner"    
}

function configure_rclone(){
  eval "$(kubectl --cache-dir /tmp -n rook-ceph exec deploy/rook-ceph-tools -- radosgw-admin user info --uid=admin  --format json | jq -r '.keys[0] | { "RCLONE_CONFIG_CEPH_ACCESS_KEY_ID" : .access_key , "RCLONE_CONFIG_CEPH_SECRET_ACCESS_KEY": .secret_key } | to_entries[] | .key + "=" + (.value | @sh)')"
  export RCLONE_CONFIG_CEPH_ACCESS_KEY_ID
  export RCLONE_CONFIG_CEPH_SECRET_ACCESS_KEY
  export RCLONE_CONFIG_CEPH_TYPE=s3
  export RCLONE_CONFIG_CEPH_PROVIDER=Ceph
  export RCLONE_CONFIG_CEPH_ENDPOINT=http://rook-ceph-rgw-rook-ceph.rook-ceph.svc
  export RCLONE_CONFIG_CEPH_REGION=us-east-1
}

function configure_s3cmd(){
  export AWS_ACCESS_KEY=$RCLONE_CONFIG_CEPH_ACCESS_KEY_ID
  export AWS_SECRET_KEY=$RCLONE_CONFIG_CEPH_SECRET_ACCESS_KEY
}

function change_bucket_owner() {
  local bucket="$1"
  local user="$2"

  kubectl --cache-dir /tmp -n rook-ceph exec deploy/rook-ceph-tools -- radosgw-admin bucket link --uid "${user}" --bucket "${bucket}" || (echo "Failure running bucket link" && return 1)
  kubectl --cache-dir /tmp -n rook-ceph exec deploy/rook-ceph-tools -- radosgw-admin bucket chown --uid "${user}" --bucket "${bucket}" || (echo "Failure running bucket chown" && return 1)
}

function validate_restore_epoch(){
  local epoch="$1"

  if ! date -d "@$epoch" >> /dev/null; then
    echo "Provide Valid value for restore epoch"
    exit 1
  fi
  if [[ ! -d "${DIFF_DIR}/${epoch}" ]]; then
    echo "Provided epoch value: '${epoch}' does not match with any of the available backup"
    find "$DIFF_DIR" -maxdepth 1 -mindepth 1 -type d -print0 | xargs -0 -I{} basename {}
    exit 1
  fi
}

function prepare_pitr_restore() {
  local bucket="$1"
  local restore_path="${RESTORE_TEMP_PATH}/${bucket}"

  rm -rf "${restore_path}"
  mkdir -p "${restore_path}"
  cp -r --preserve=timestamps "${DST_DIR}/${bucket}/." "${restore_path}"

  for diff_dir in $(find "$DIFF_DIR" -maxdepth 1 -mindepth 1  -type d -newermt "@${RESTORE_EPOCH}"  | sort -r | sed '$ d'); do
    if [[ -d "${diff_dir}/${bucket}" ]]; then
      cp -r --preserve=timestamps "${diff_dir}/${bucket}/." "${restore_path}"
    fi
  done

  find "${restore_path}" -newermt "@${RESTORE_EPOCH}" -type f -print0 | xargs -0 -r rm -f
}

function main(){
  EPOCH=$(date +"%s")
  BACKUP_DIR="${NFS_PATH}/backup/s3"
  LOG_DIR="${BACKUP_DIR}/logs"
  LOG_FILE="${LOG_DIR}/${EPOCH}/backup.log"
  DIFF_DIR="${BACKUP_DIR}/diff"
  DST_DIR="${BACKUP_DIR}/latest"
  EXCLUDE_BUCKETS_ARRAY=$(jq -c -n --arg buckets "${EXCLUDE_BUCKETS}" '[$buckets | split(",") | .[] | gsub("^\\s+|\\s+$";"")]')
  INCLUDE_BUCKETS_ARRAY=$(jq -c -n --arg buckets "${INCLUDE_BUCKETS}" '[$buckets | split(",") | .[] | gsub("^\\s+|\\s+$";"")]')
  configure_rclone
  configure_s3cmd
  case "$OP" in 
    backup)
      buckets=$(kubectl --cache-dir /tmp -n rook-ceph exec deploy/rook-ceph-tools -- radosgw-admin bucket list  --format json | jq -r  '.[]')
      ;;
    restore)
      [[ $RESTORE_EPOCH != "latest" ]] && validate_restore_epoch "$RESTORE_EPOCH"
      if [[ -d $DST_DIR ]]; then
        buckets=$(find "$DST_DIR" -maxdepth 1 -mindepth 1 -type d -print0 | xargs -0 -I{} basename {})
        if [[ -n $buckets ]]; then
          LOG_FILE="${NFS_PATH}/restore/s3/logs/${EPOCH}/restore.log"
        fi
      else
        echo "Backup path not found at '${DST_DIR}'"
        exit 1
      fi
      ;;
    *)
      echo "Unsupported value: '$OP' for operation"
      exit 1
      ;;
  esac

  if [[ -z $buckets ]]; then
    echo "Bucket list is empty, cannot perform ${OP}"
    exit 1
  fi

  mkdir -p "$(dirname "${LOG_FILE}")"
  [[ $OP == "backup" ]] && mkdir -p "${DIFF_DIR}/${EPOCH}"
  echo "Using file: '${LOG_FILE}' for storing logs"
  tail -F "${LOG_FILE}" 2> /dev/null &
  logs_tail_pid="$!"
  op_pid=()
  for bucket in $buckets; do
    include_bucket='false'
    if [[ -n $INCLUDE_BUCKETS_PATTERN ]] && echo "${bucket}" | grep -q -P "$INCLUDE_BUCKETS_PATTERN"; then
      include_bucket='true'
    fi

    if [[ $INCLUDE_BUCKETS_ARRAY != '[]' ]]; then
      if ! jq -n --argjson includeList "${INCLUDE_BUCKETS_ARRAY}" --arg bucket "$bucket" '$includeList | index($bucket)' | grep -q 'null'; then
        include_bucket='true'
      fi
    fi

    if [[ -n $INCLUDE_BUCKETS_PATTERN || $INCLUDE_BUCKETS_ARRAY != '[]' ]] && [[ $include_bucket == "false" ]]; then
      echo "Skipping ${OP} for bucket: '${bucket}' Not Found in include list or include pattern"
      continue
    fi

    if ! jq -n --argjson excludeList "${EXCLUDE_BUCKETS_ARRAY}" --arg bucket "$bucket" '$excludeList | index($bucket)' | grep -q 'null'; then
      echo "Skipping ${OP} for bucket: '${bucket}' Found in exclude list"
      continue
    fi
    if [[ -n $EXCLUDE_BUCKETS_PATTERN ]] && echo "${bucket}" | grep -q -P "$EXCLUDE_BUCKETS_PATTERN"; then
      echo "Skipping ${OP} for bucket: '${bucket}' Matching exclude pattern"
      continue
    fi

    echo "Running ${OP} for bucket: '${bucket}'"
    owner_file="${BACKUP_DIR}/info/${bucket}/owner.txt"
    cors_file="${BACKUP_DIR}/info/${bucket}/cors.xml"
    policy_file="${BACKUP_DIR}/info/${bucket}/policy.json"
    acl_file="${BACKUP_DIR}/info/${bucket}/acl.xml"
    lifecycle_file="${BACKUP_DIR}/info/${bucket}/lifecycle.xml"
    if [[ $OP == "backup" ]]; then
      mkdir -p "$(dirname "${owner_file}")"

      bucket_owner=$(get_bucket_owner "${bucket}")
      [[ -n "$bucket_owner" ]] && echo "${bucket_owner}" > "$owner_file"
      [[ -z "$bucket_owner" ]] && echo "Unable to get owner of the bucket: '${bucket}'"

      bucket_cors=$(get_bucket_config "${bucket}" "cors")
      [[ -n "$bucket_cors" ]] && echo "${bucket_cors}" > "$cors_file"
      [[ -z "$bucket_cors" ]] && echo "No CORS found for bucket: '${bucket}'"

      bucket_policy=$(get_bucket_config "${bucket}" "policy")
      [[ -n "$bucket_policy" ]] && echo "${bucket_policy}" > "$policy_file"
      [[ -z "$bucket_policy" ]] && echo "No bucket policy found for bucket: '${bucket}'" 
      
      bucket_acl=$(get_bucket_config "${bucket}" "acl")
      [[ -n "$bucket_acl" ]] && echo "${bucket_acl}" > "$acl_file"
      [[ -z "$bucket_acl" ]] && echo "No ACL found for bucket: '${bucket}'"

      bucket_lifecycle=$(get_bucket_config "${bucket}" "lifecycle")
      [[ -n "$bucket_lifecycle" ]] && echo "${bucket_lifecycle}" > "$lifecycle_file"
      [[ -z "$bucket_lifecycle" ]] && echo "No Lifecycle policy found for bucket: '${bucket}'"
      # Actual Backup step
      rclone sync "ceph:${bucket}" "${DST_DIR}/${bucket}"  --stats=0 --backup-dir "${DIFF_DIR}/${EPOCH}/${bucket}"  --log-level INFO --log-file "$LOG_FILE" &
      op_pid+=("${bucket}=$!")  
    fi
    
    if [[ $OP == "restore" ]]; then
      restore_src="${DST_DIR}/${bucket}"
      if [[ $RESTORE_EPOCH != "latest" ]]; then
        prepare_pitr_restore "${bucket}"
        restore_src="${RESTORE_TEMP_PATH}/${bucket}"
      fi
      rclone sync "${restore_src}" "ceph:${bucket}" --stats=0 --log-level INFO --log-file "$LOG_FILE" &
      op_pid+=("${bucket}=$!")
    fi
  done

  exit_code=0
  for pid in "${op_pid[@]}"; do
    bucket=$(echo "$pid" | cut -d'=' -f1)
    cpid=$(echo "$pid" | cut -d'=' -f2)
    owner_file="${BACKUP_DIR}/info/${bucket}/owner.txt"
    cors_file="${BACKUP_DIR}/info/${bucket}/cors.xml"
    policy_file="${BACKUP_DIR}/info/${bucket}/policy.json"
    acl_file="${BACKUP_DIR}/info/${bucket}/acl.xml"
    lifecycle_file="${BACKUP_DIR}/info/${bucket}/lifecycle.xml"
    if ! wait "$cpid"; then
      echo "Failure running ${OP} for bucket '${bucket}'"
      exit_code=1
    else
      if [[ $OP == "restore" ]]; then
        ## Restore bucket owner
        owner=""
        [[ -f "$owner_file" && -s "$owner_file" ]] && owner=$(cat "$owner_file")
        if [[ -n "${owner}" ]]; then
          echo "Restoring bucket: '${bucket}' owner to: '${owner}'"
          if ! change_bucket_owner "$bucket" "$owner"; then
            echo "Failure while restoring bucket: '${bucket}' owner to: '${owner}'"
            exit_code=1
          else
            echo "Successfully restored bucket: '${bucket}' owner to: '${owner}'"
          fi
        else
          echo "Skipping owner restore for bucket: '${bucket}' as owner not found in backup"
        fi

         # Restore ACL
        if [[ -f "$acl_file" && -s "$acl_file" ]]; then
          restore_bucket_config "${bucket}" "acl" "$acl_file" || exit_code=1
        else
          echo "Skipping ACL restore for bucket: '${bucket}' as ACL config not found in backup"
        fi
         # Restore CORS
         if [[ -f "$cors_file" && -s "$cors_file" ]]; then
          restore_bucket_config "${bucket}" "cors" "$cors_file" || exit_code=1
        else
          echo "Skipping CORS restore for bucket: '${bucket}' as CORS config not found in backup"
        fi

         # Restore bucket policy
        if [[ -f "$policy_file" && -s "$policy_file" ]]; then
          restore_bucket_config "${bucket}" "policy" "$policy_file" || exit_code=1
        else
          echo "Skipping Policy restore for bucket: '${bucket}' as Policy config not found in backup"
        fi

         # Restore lifecycle policy
        if [[ -f "$lifecycle_file" && -s "$lifecycle_file" ]]; then
          restore_bucket_config "${bucket}" "lifecycle" "$lifecycle_file" || exit_code=1
        else
          echo "Skipping Lifecycle Policy restore for bucket: '${bucket}' as Lifecycle Policy config not found in backup"
        fi
      fi
    fi
  done

  [[ $OP == "restore" ]] && rm -rf "${RESTORE_TEMP_PATH}"
  log_epoch=$(date -d "$LOG_RETENTION_DAYS days ago" +"%s")
  diff_epoch=$(date -d "$DIFF_RETENTION_DAYS days ago" +"%s")
  [[ $exit_code -eq 0 && $OP == "backup" ]] && clean_up_old_dir "$log_epoch" "$LOG_DIR"
  [[ $exit_code -eq 0 && $OP == "backup" ]] && clean_up_old_dir "$diff_epoch" "$DIFF_DIR"
  kill -9 "$logs_tail_pid"
  wait "$logs_tail_pid" 2>/dev/null || true
  [[ $exit_code -eq 0 ]] && echo "$(date +'%Y-%m-%d %H:%M:%S') S3 bucket ${OP} completed successfully"
  return $exit_code
}

main