#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

set -eu -o pipefail

#GLOBALS
MODULES_DIR=$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)
if [[ -z "${MODULES_DIR}" ]]; then
  error "Could not determine modules path"
fi

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/cluster_utils.sh

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/utils.sh

#shellcheck disable=SC1090,SC1091
source "${MODULES_DIR}"/offline-utils.sh


ROOT_DIR=$(cd "${MODULES_DIR}/.." && pwd)
if [[ -z "${ROOT_DIR}" ]]; then
  error "Could not determine root path"
fi

BASE_NAME=$(basename "$0")
COMMAND_NAME="${COMMAND_NAME:-$BASE_NAME}"

readonly MODULES_DIR

export MODULES_DIR
export KUBECONFIG=${KUBECONFIG:-/etc/rancher/rke2/rke2.yaml}
export PATH=$PATH:/var/lib/rancher/rke2/bin:${ROOT_DIR}/Fabric_Installer/bin:/usr/local/bin
export OFFLINE_TMP_FOLDER="/tmp"


function display_usage() {
  echo
  echo "***************************************************************************************"
  echo
  echo "Upload optional offline bundle"
  echo
  echo "Usage:"
  echo "  $COMMAND_NAME [flags]"
  echo
  echo "Flags:"
  echo "  --optional-offline-bundle           Specify the path to the optional offline bundle"
  echo "  --offline-bundle                    Specify the path to the offline bundle"
  echo "  --offline-tmp-folder                Specify parent directory for the temporary offline install folder"
  echo "  -h|--help                           Display help"
  echo
  echo "***************************************************************************************"
  echo
}

function parse_long_args() {
  while (("$#")); do
    case "$1" in
    -d | --debug)
      info "Running in debug"
      shift
      set -x
      ;;

    -h | --help)
      display_usage
      shift
      exit 0
      ;;

    --optional-offline-bundle)
      if [ -n "${2:-}" ] && [ "${2:0:1}" != "-" ]; then
        local absolute_path
        absolute_path=$(cd "${ROOT_DIR}" && readlink -f "$2") || error_with_code "10032" "$2"
        OPTIONAL_OFFLINE_BUNDLE_PATH="${absolute_path}"
        shift 2
      else
        display_usage
        error_with_code "10005" "$1"
        shift 2
      fi
      ;;

    --offline-bundle)
      if [ -n "${2:-}" ] && [ "${2:0:1}" != "-" ]; then
        local absolute_path
        absolute_path=$(cd "${ROOT_DIR}" && readlink -f "$2") || error_with_code "10032" "$2"
        OFFLINE_BUNDLE_PATH="${absolute_path}"
        shift 2
      else
        display_usage
        error_with_code "10005" "$1"
        shift 2
      fi
      ;;

    --offline-tmp-folder)
      if [ -n "${2:-}" ] && [ "${2:0:1}" != "-" ]; then
        local absolute_path
        absolute_path=$(cd "${ROOT_DIR}" && readlink -f "$2") || error_with_code "10032" "$2"
        OFFLINE_TMP_FOLDER="${absolute_path}"
        shift 2
      else
        display_usage
        error_with_code "10005" "$1"
        shift 2
      fi
      ;;

    -*|*) # unsupported flags
      display_usage
      exit 1
      ;;
    esac
  done
}

function set_offline_variables() {
  local docker_registry_password_name="docker-registry-credentials"
  local cluster_port
  local registry_fqdn
  registry_fqdn=$(get_cluster_config "registry_fqdn")
  cluster_port=$(get_cluster_config "cluster_port")
  HELM_REGISTRY_URL="${registry_fqdn}:${cluster_port}"
  DOCKER_REGISTRY_NODEPORT=$(get_cluster_config "registry_nodeport")
  PODMAN_FOLDER=$(get_cluster_config "podman_folder")
  HELM_REGISTRY_PASSWORD=$(kubectl get secrets/"${docker_registry_password_name}" -n "uipath" --template \{\{.data.password\}\} | base64 -d)
  HELM_REGISTRY_USERNAME=$(kubectl get secrets/"${docker_registry_password_name}" -n "uipath" --template \{\{.data.username\}\} | base64 -d)
  export HELM_REGISTRY_URL
  export HELM_REGISTRY_PASSWORD
  export HELM_REGISTRY_USERNAME
  export DOCKER_REGISTRY_NODEPORT=${DOCKER_REGISTRY_NODEPORT:-"30071"}
  export PODMAN_FOLDER
}


#####
# $1 - offline bundle path
# can be either optional or the main artifact archive
# $2 - temporary folder path
#####
function prepare_bundle_load() {
  local offline_bundle_path="${1}"
  local tmp_offline_folder="${2}"
  info "Extracting ${offline_bundle_path} to ${tmp_offline_folder}"
  tar -xzf "${offline_bundle_path}" -C "${tmp_offline_folder}"
  info "Files used for the offline bundle have been extracted to ${tmp_offline_folder}"
  set_offline_variables
}

function load_optional_offline_bundle() {
  info "Importing optional bundle at ${OPTIONAL_OFFLINE_BUNDLE_PATH}"
  local tmp_offline_folder
  tmp_offline_folder="$(mktemp -d -p "${OFFLINE_TMP_FOLDER}" -t offline-bundle.XXXXXXXXXX)"
  prepare_bundle_load "${OPTIONAL_OFFLINE_BUNDLE_PATH}" "${tmp_offline_folder}"
  import_on_demand_artifacts "${PODMAN_FOLDER}" "${tmp_offline_folder}"
}

function load_offline_bundle() {
  info "Importing bundle at ${OFFLINE_BUNDLE_PATH}"
  local tmp_offline_folder
  tmp_offline_folder="$(mktemp -d -p "${OFFLINE_TMP_FOLDER}" -t offline-bundle.XXXXXXXXXX)"
  prepare_bundle_load "${OFFLINE_BUNDLE_PATH}" "${tmp_offline_folder}"
  import_offline_sf_artifacts "${PODMAN_FOLDER}" "${tmp_offline_folder}"
}

function validate_parameters() {
  if [[ -z "${OPTIONAL_OFFLINE_BUNDLE_PATH:-}" ]] && [[ -z "${OFFLINE_BUNDLE_PATH:-}" ]]; then
    display_usage
    error_with_code "10033"
  fi
  if [[ -n ${OFFLINE_TMP_FOLDER:-} ]] && [[ ! -d "${OFFLINE_TMP_FOLDER}" ]]; then
    mkdir -p "${OFFLINE_TMP_FOLDER}"
  fi
}

function main() {
  parse_long_args "$@"
  validate_parameters
  if [[ -z "${OPTIONAL_OFFLINE_BUNDLE_PATH:-}" ]]; then
    load_offline_bundle
  else
    load_optional_offline_bundle
  fi
}

main "$@"
exit 0
