#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

#GLOBALS
export PATH="$PATH:/usr/local/sbin:/sbin:/bin:/usr/sbin:/usr/bin:/root/bin:/usr/local/bin"

SCRIPT_DIR=$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)
if [[ -z "${SCRIPT_DIR}" ]]; then
  echo "Could not determine script path"
  exit 1
fi

MESSAGE_FILE="${SCRIPT_DIR}/../readiness_message_en.json"
if [[ ! -f "${MESSAGE_FILE}" ]] ; then
  echo "Readiness message file not present"
  exit 1
fi

MODULES_DIR=$(cd "${SCRIPT_DIR}/../../Modules" && pwd)
if [[ -z "${MODULES_DIR}" ]]; then
  echo "Could not determine Modules path"
  exit 1
fi

export MODULES_DIR

if [[ -f "${MODULES_DIR}/utils.sh" ]] ; then
  #shellcheck disable=SC1091,SC1090
  source "${MODULES_DIR}/utils.sh"
else
  echo "Common utils not found at ${MODULES_DIR}/utils.sh"
  exit 1
fi

EXIT_CODE=0

function display_usage() {
  echo
  echo "***************************************************************************************"
  echo
  echo "Run certificates checks"
  echo
  echo "Arguments"
  echo "  -d|--debug                         Increase logging verbosity to show all debug logs"
  echo "  -h|--help                          Show this help message and exit"
  echo  
  echo "***************************************************************************************"
  echo
}

function parse_long_args() {
  while (("$#")); do
    case "$1" in
    -d | --debug)
      echo "Running in debug"
      shift
      set -x
      ;;

    -h | --help)
      display_usage
      shift
      exit 0
      ;;

    -*|*) # unsupported flags
      # display_usage
      # echo "Error: Unsupported flag $1"
      # exit 1
      shift
      ;;
    esac
  done
}

function check_networkmanager_services() {
    if systemctl is-enabled nm-cloud-setup.service &>/dev/null ;
    then
      add_test_result "FAILED" "$(get_readiness_message ".system_service.nm_cloud_service_config_error" "${MESSAGE_FILE}" )" "$(get_readiness_mitigation ".system_service.nm_cloud_service_config_error" "${MESSAGE_FILE}" )"
    else
      add_test_result "PASSED" "$(get_readiness_message ".system_service.nm_cloud_service_config_info" "${MESSAGE_FILE}" )" "$(get_readiness_mitigation ".system_service.nm_cloud_service_config_info" "${MESSAGE_FILE}" )"
    fi

    if systemctl is-enabled nm-cloud-setup.timer &>/dev/null ;
    then
      add_test_result "FAILED" "$(get_readiness_message ".system_service.nm_cloud_timer_config_error" "${MESSAGE_FILE}" )" "$(get_readiness_mitigation ".system_service.nm_cloud_timer_config_error" "${MESSAGE_FILE}" )"
    else
      add_test_result "PASSED" "$(get_readiness_message ".system_service.nm_cloud_timer_config_info" "${MESSAGE_FILE}" )" "$(get_readiness_mitigation ".system_service.nm_cloud_timer_config_info" "${MESSAGE_FILE}" )"
    fi
}

function validate_auditd_state() {
    if systemctl is-enabled auditd &>/dev/null;
    then
        # enabled set to 2 means, it is immutable
        if [[ "$(get_auditctl_parameter_value "enabled")" == "2" ]] && [[ "$(get_auditctl_parameter_value "backlog_wait_time")" != "1" ]]; then
          add_test_result "FAILED" "$(get_readiness_message ".system_service.auditd_config_error" "${MESSAGE_FILE}" )" "$(get_readiness_mitigation ".system_service.auditd_config_error" "${MESSAGE_FILE}" )"
        elif [[ "$(get_auditctl_parameter_value "backlog_wait_time")" != "1" ]]; then
          add_test_result "WARN" "$(get_readiness_message ".system_service.auditd_config_warn" "${MESSAGE_FILE}" )" "$(get_readiness_mitigation ".system_service.auditd_config_warn" "${MESSAGE_FILE}" )"
        fi
    fi
}

function check_iscsid_status() {
    if ! systemctl is-active iscsid &>/dev/null ;
    then
      add_test_result "WARN" "$(get_readiness_message ".system_service.iscsid_config_warn" "${MESSAGE_FILE}" )" "$(get_readiness_mitigation ".system_service.iscsid_config_warn" "${MESSAGE_FILE}" )"
    else
      add_test_result "PASSED" "$(get_readiness_message ".system_service.iscsid_config_info" "${MESSAGE_FILE}" )" "$(get_readiness_mitigation ".system_service.iscsid_config_info" "${MESSAGE_FILE}" )"
    fi
}

function check_firewalld_status() {
  if systemctl is-enabled firewalld &>/dev/null;
  then
    add_test_result "WARN" "$(get_readiness_message ".system_service.firewalld_config_warn" "${MESSAGE_FILE}" )" "$(get_readiness_mitigation ".system_service.firewalld_config_warn" "${MESSAGE_FILE}" )"
  else
    add_test_result "PASSED" "$(get_readiness_message ".system_service.firewalld_config_info" "${MESSAGE_FILE}" )" "$(get_readiness_mitigation ".system_service.firewalld_config_info" "${MESSAGE_FILE}" )"
  fi
}

function check_ipforwarding_status() {
  if [[ $(cat /proc/sys/net/ipv4/ip_forward) != "1" ]]; then
    add_test_result "WARN" "$(get_readiness_message ".system_service.ipv4_forwarding_config_warn" "${MESSAGE_FILE}" )" "$(get_readiness_mitigation ".system_service.ipv4_forwarding_config_warn" "${MESSAGE_FILE}" )"
  else
    add_test_result "PASSED" "$(get_readiness_message ".system_service.ipv4_forwarding_config_info" "${MESSAGE_FILE}" )" "$(get_readiness_mitigation ".system_service.ipv4_forwarding_config_info" "${MESSAGE_FILE}" )"
  fi
}

function grep_keyword_in_logrotate_dir() {
    local grep_keyword="$1"
    local logrotate_dir_path
    logrotate_dir_path="/etc/logrotate.d/"
    if [[ -d "${logrotate_dir_path}" ]]; then
        mapfile -t files < <(grep -rl "${logrotate_dir_path}" -e "${grep_keyword}")
        for file in "${files[@]}"; do
            error_messages+=("$(get_readiness_message ".system_service.logrotate_config_warn_message" "${MESSAGE_FILE}" "${grep_keyword}" "${file}" )")
        done
    fi
}

function validate_logrotated_config() {
    echo -e "$(get_readiness_message ".system_service.logrotate_info" "${MESSAGE_FILE}" )"
    error_messages=()
    grep_keyword_in_logrotate_dir "weekly"
    grep_keyword_in_logrotate_dir "monthly"
    grep_keyword_in_logrotate_dir "yearly"
    if [[ "${#error_messages[@]}" != 0 ]];
    then
      add_test_result "WARN" "$(get_readiness_message ".system_service.logrotate_config_warn_title" "${MESSAGE_FILE}")" "$(get_mitigation_string_from_array "${error_messages[@]}")"
    else
      add_test_result "PASSED" "$(get_readiness_message ".system_service.logrotate_config_pass" "${MESSAGE_FILE}")"
    fi
    echo -e "$(get_readiness_message ".system_service.logrotate_end" "${MESSAGE_FILE}" )\n"
}

function systemd_services_validation() {
    #shellcheck disable=SC2005
    echo "$(get_readiness_message ".system_service.start_info" "${MESSAGE_FILE}" )"
    check_firewalld_status

    check_networkmanager_services

    check_ipforwarding_status

    validate_auditd_state

    check_iscsid_status
    echo -e "$(get_readiness_message ".system_service.end_info" "${MESSAGE_FILE}" )\n"
    
    print_center ""
    
    print_center "Log Rotation Config"
    validate_logrotated_config
}

function main() {
  parse_long_args "$@"
  print_center "Systemd Services Validation"
  systemd_services_validation
  print_center ""
}

main "$@"
exit "${EXIT_CODE}"