[CmdletBinding()]
    param(
        [Parameter(Mandatory = $true)]
        [string] $TestRailBranch,

        [Parameter(Mandatory = $true)]
        $TestRailUserName,

        [Parameter(Mandatory = $true)]
        $TestRailPass,

        [Parameter(Mandatory = $true)]
        [string] $TestRunName,

        [Parameter(Mandatory = $true)]
        $TestRailProjectID,

        [Parameter(Mandatory = $true)]
        $TestRailSuiteID,

        [Parameter(Mandatory = $true)]
        $TestRailPipelineToken,

        [Parameter(Mandatory = $false)]
        $TestRailCompanyDomain = "uipath.testrail.com",

        [Parameter(Mandatory = $false)]
        $AzureDevopsUri = "dev.azure.com",

        [Parameter(Mandatory = $false)]
        $AzureDevOpsOrganisation = "uipath",

        [Parameter(Mandatory = $false)]
        $AzureDevOpsProject = "Service Fabric",

        [Parameter(Mandatory = $false)]
        $TestRailPipelineId = 2180
    )

    $ErrorActionPreference = "Continue"

    function New-TestRun {

        $testRunDetails = @{
            suite_id=$TestRailSuiteID
            name=$TestRunName
            include_all=$true
        } | ConvertTo-Json
    
        $base64AuthInfo = [Convert]::ToBase64String([Text.Encoding]::ASCII.GetBytes(("{0}:{1}" -f $TestRailUserName, $TestRailPass)))
        $headers = @{
            Authorization=("Basic {0}" -f $base64AuthInfo)
            custom_automation_type="SF-pipeline"
            "Content-Type"="application/json"
        }

        $uri = New-Object System.UriBuilder -argumentlist 'https', $TestRailCompanyDomain, 443, '/index.php', "?/api/v2/add_run/$TestRailProjectID"
        $testRun = Invoke-RestMethod -Method POST -Uri $uri.Uri.AbsoluteUri -Headers $headers -Body $testRunDetails -UseBasicParsing -TimeoutSec 120
    
        return $testRun
    }

    function New-PipelineRun {
        [CmdletBinding()]
        param(
            [Parameter(Mandatory)]
            [string] $TestRailBranch,
            [Parameter(Mandatory)]
            [string] $TestRailPipelineToken,
            [Parameter(Mandatory)]
            [string] $TestRunId,
            [Parameter(Mandatory)]
            [string] $TestRailTags,
            [Parameter(Mandatory)]
            [bool] $TaskMiningAgent,
            [Parameter(Mandatory)]
            [int] $NodeCount,
            [Parameter(Mandatory)]
            [int] $ServerNodeCount,
            [Parameter(Mandatory)]
            [string] $Location
        ) 
        
        $pipelineParameters = @{
            testRailTags=$TestRailTags
            taskMiningAgent=$TaskMiningAgent
            nodeCount=$NodeCount
            serverNodeCount=$ServerNodeCount
            testRailRunId=$TestRunId
            location=$Location
        }

        $pipelineBranch = @{
            refName="$TestRailBranch"
        }

        $pipelineBranchDetails = @{
            self=$pipelineBranch
        }
        $pipelineResources = @{
            repositories=$pipelineBranchDetails
        }

        $pipelineRunDetails = @{
            templateParameters=$pipelineParameters
            resources=$pipelineResources
        } | ConvertTo-Json -Depth 100

        $base64AuthInfo = [Convert]::ToBase64String([Text.Encoding]::ASCII.GetBytes(("{0}:{1}" -f "testrailnightlybuild", $TestRailPipelineToken)))
        $headers=@{
            "Content-Type"="application/json"
            Authorization=("Basic {0}" -f $base64AuthInfo)
        }
        $uri = New-Object System.UriBuilder -argumentlist 'https' , $AzureDevopsUri, 443, "/$AzureDevOpsOrganisation/$AzureDevOpsProject/_apis/pipelines/$TestRailPipelineId/runs","?api-version=6.0-preview.1"

        try {
            $pipelineResponse = Invoke-RestMethod -Method POST -Uri $uri.Uri.AbsoluteUri -Body $pipelineRunDetails -Headers $headers -UseBasicParsing -TimeoutSec 120
            return $pipelineResponse
        } catch {
            # Dig into the exception to get the Response details.
            # Note that value__ is not a typo.
            $StatusCode=$_.Exception.Response.StatusCode.value__ 
            $StatusDescription=$_.Exception.Response.StatusDescription
            Write-Host "Status Code: $StatusCode, Status Description: $StatusDescription"
            Write-Error "Failed to Create Pipeline."
        }
    }

    function Get-PipelineStatus(){
        [CmdletBinding()]
        param(
            [Parameter(Mandatory)]
            [string] $PipeineUrl,
            [Parameter(Mandatory)]
            $TestRailPipelineToken
        ) 

        $base64AuthInfo = [Convert]::ToBase64String([Text.Encoding]::ASCII.GetBytes(("{0}:{1}" -f "testrailnightlybuild", $TestRailPipelineToken)))
        $headers=@{
            "Content-Type"="application/json"
            Authorization=("Basic {0}" -f $base64AuthInfo)
        }

        try {
            $pipelineResponse = Invoke-RestMethod -Method GET -Uri $PipeineUrl -Headers $headers -UseBasicParsing -TimeoutSec 120
            return $pipelineResponse
        } catch {
            # Dig into the exception to get the Response details.
            # Note that value__ is not a typo.
            $StatusCode=$_.Exception.Response.StatusCode.value__ 
            $StatusDescription=$_.Exception.Response.StatusDescription
            Write-Host "Status Code: $StatusCode, Status Description: $StatusDescription"
        }

    }

    function Execute-Pipelines {
        [CmdletBinding()]
        param(
            [Parameter(Mandatory)]
            [System.Object[]] $TestRun,
            [Parameter(Mandatory)]
            $TestRailPipelineToken,
            [Parameter(Mandatory)]
            [string] $TestRailBranch
        )

        # Trigger Pipeline
        $TestRunId=$TestRun.id

        $pipelineResponse1 = New-PipelineRun -TestRailBranch $TestRailBranch -TestRailPipelineToken $TestRailPipelineToken -TestRunId $TestRunId -TestRailTags "Readiness,Installation,SupportBundle,Alerts" -TaskMiningAgent $false -NodeCount 1 -ServerNodeCount 1 -Location "westeurope"
        $pipelineResponse2 = New-PipelineRun -TestRailBranch $TestRailBranch -TestRailPipelineToken $TestRailPipelineToken -TestRunId $TestRunId -TestRailTags "CertManagement,PVCResize,ClusterConfiguration" -TaskMiningAgent $false -NodeCount 1 -ServerNodeCount 1 -Location "northeurope"
        $pipelineResponse3 = New-PipelineRun -TestRailBranch $TestRailBranch -TestRailPipelineToken $TestRailPipelineToken -TestRunId $TestRunId -TestRailTags "HighAvailability,AgentTypes" -TaskMiningAgent $true -NodeCount 3 -ServerNodeCount 3 -Location "centralus"

        while ($true) {
            Start-Sleep 120

            $pipelineStatusResponse1 = Get-PipelineStatus -PipeineUrl $pipelineResponse1.url -TestRailPipelineToken $TestRailPipelineToken
            $pipelineStatusResponse2 = Get-PipelineStatus -PipeineUrl $pipelineResponse2.url -TestRailPipelineToken $TestRailPipelineToken
            $pipelineStatusResponse3 = Get-PipelineStatus -PipeineUrl $pipelineResponse3.url -TestRailPipelineToken $TestRailPipelineToken

            if ($pipelineStatusResponse1.state -ne "inProgress" -and $pipelineStatusResponse2.state -ne "inProgress" -and $pipelineStatusResponse3.state -ne "inProgress") {
                Write-Host "$pipelineStatusResponse1"
                Write-Host "$pipelineStatusResponse2"
                Write-Host "$pipelineStatusResponse3"
                if ($pipelineStatusResponse1.result -eq "failed" -or $pipelineStatusResponse2.result -eq "failed" -or $pipelineStatusResponse3.result -eq "failed"){
                    Write-Error "Pipeline Failed."
                }
                else {
                    Write-Host "Pipeline completed."
                }
                break;
            }
            else {
                Write-Host "$pipelineStatusResponse1"
                Write-Host "$pipelineStatusResponse2"
                Write-Host "$pipelineStatusResponse3"
                continue;
            }
        }
    }
    
    function Main {    
        $testRun = New-TestRun
        Execute-Pipelines -TestRun $testRun -TestRailPipelineToken $TestRailPipelineToken -TestRailBranch $TestRailBranch
    }
    
Main
Exit 0