#!/bin/bash

# Copyright UiPath 2021
#
# =================
# LICENSE AGREEMENT
# -----------------
#   Use of paid UiPath products and services is subject to the licensing agreement
#   executed between you and UiPath. Unless otherwise indicated by UiPath, use of free
#   UiPath products is subject to the associated licensing agreement available here:
#   https://www.uipath.com/legal/trust-and-security/legal-terms (or successor website).
#   You must not use this file separately from the product it is a part of or is associated with.

#GLOBALS
set -e
SCRIPT_DIR=$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)
if [[ -z "${SCRIPT_DIR}" ]]; then
  echo "Could not determine script path"
  exit 1
fi

MODULES_DIR=$(cd "${SCRIPT_DIR}/../Modules" && pwd)
if [[ -z "${MODULES_DIR}" ]]; then
  echo "Could not determine Modules path"
  exit 1
fi

if [[ -f "${MODULES_DIR}/utils.sh" ]] ; then
  #shellcheck disable=SC1091,SC1090
  source "${MODULES_DIR}/utils.sh"
else
  echo "Common utils not found at ${MODULES_DIR}/utils.sh"
  exit 1
fi

if [[ ! -f "${SCRIPT_DIR}"/../configureUiPathAS.sh ]]; then
  error "configureUiPathAS.sh not found at ${SCRIPT_DIR}/../configureUiPathAS.sh"
fi

if [[ ! -x "${SCRIPT_DIR}"/../configureUiPathAS.sh ]] || ! "${SCRIPT_DIR}"/../configureUiPathAS.sh -h >> /dev/null 2>&1 ; then
  error "${SCRIPT_DIR}/../configureUiPathAS.sh does not have executable permission"
fi


LONGHORN_NAMESPACE="longhorn-system"
LONGHORN_MANGER_LABELS="app=longhorn-manager"
LH_DATADISK_PATH="/datadisk"

ROOK_CEPH_NAMESPACE="rook-ceph"
ROOK_CEPH_TOOLS_CONTROLLER="deploy/rook-ceph-tools"
ROOK_CEPH_DATA_POOL_NAME="rook-ceph.rgw.buckets.data"
ROOK_CEPH_EC_PROFILE_NAME="rook-ceph_ecprofile"
SCALE_UP_ROOK_CEPH_OPERATOR="false"
ENABLE_CEPH_TRAFFIC="false"
OBJECT_COUNT=""
HARD_LOCK_FILE="${SCRIPT_DIR}/ceph-migration.lock"
GLOBAL_FORCE=""
EC_DATA_POOL_NAME="ec-data-pool"
DELETE_TEMP_EC_POOL="false"


SKIP_CHECK_ON_EXIT="true"
FAILURE_ON_EXIT="false"

export KUBECONFIG="/etc/rancher/rke2/rke2.yaml"
export PATH="$PATH:/usr/local/bin:/var/lib/rancher/rke2/bin"


LOG_FILE="/tmp/cppool.log"
trap on_exit 0

function on_exit() {
  [[ $SKIP_CHECK_ON_EXIT == "true" ]] && return
  [[ $DELETE_TEMP_EC_POOL == "true" ]] && delete_data_pool "${EC_DATA_POOL_NAME}"
  [[ $ENABLE_CEPH_TRAFFIC == "true" ]] && enable_disable_ceph_traffic "enable"
  [[ $SCALE_UP_ROOK_CEPH_OPERATOR == "true" ]] && scale_rook_ceph_operator "1"
  if [[ $FAILURE_ON_EXIT == "true" ]] ; then
    error "Ceph data pool migration failed"
  else
    info "Ceph data pool migration completed successfully"
  fi
}

function delete_crashloop_backoff_pods(){
  #shellcheck disable=SC2016
  kubectl -n uipath get pod -A --no-headers  | grep 'CrashLoopBackOff' -i  | awk '{print $1,$2}' | xargs -L1  bash -c 'kubectl -n "$1" delete pod --grace-period=0 --force "$2"' "$0" || true
}

function check_hard_lock() {
  if [[ -f "${HARD_LOCK_FILE}" ]]; then
    error "Running migration tool after an fatal error is not supported as it may lead to data loss. Raise a support ticket with uipath to complete the migration"
  fi
}

function create_hard_lock(){
  touch "$HARD_LOCK_FILE"
  [[ -f "$HARD_LOCK_FILE" ]] || warn "Failure while creating lock file"
  chattr +i +a "$HARD_LOCK_FILE" >> /dev/null 2>&1 || true
}

function display_usage() {
  echo
  echo "***************************************************************************************"
  echo
  echo "Migrate Ceph data pool from replicated to erasure coded pool"
  echo
  echo "Arguments"
  echo "  -l|--log                            Log file path to store actual ceph pool migration logs"
  echo "  --force                             Provide consent to perform action which may cause downtime to complete migration"
  echo "  -h|--help                           Show this help message and exit"
  echo
  echo "***************************************************************************************"  
  echo
}

function parse_long_args() {
  while (("$#")); do
    case "$1" in
    -l | --log)
      if [ -n "$2" ] && [ "${2:0:1}" != "-" ]; then
        LOG_FILE="$2"
        touch "$LOG_FILE" >> /dev/null 2>&1 || error "Provided log file path: '$(dirname "$2")' does not exist or User does not permission to write log file. Please provide correct path"
        \rm -rf "$LOG_FILE"
        shift 2
      else
        error "Value for $1 is missing"
        shift 2
      fi
      ;;

    -f | --force)
      warn "Running with force=true"
      shift
      GLOBAL_FORCE="true"
      ;;

    -h | --help)
      display_usage
      shift
      exit 0
      ;;

    -*|*) # unsupported flags
      display_usage
      error "Error: Unsupported flag $1"
      exit 1
      ;;
    esac
  done
}

function run_ceph_cmd() {
  local cmd="$1"
  #shellcheck disable=SC2086
  kubectl -n "${ROOK_CEPH_NAMESPACE}" exec "${ROOK_CEPH_TOOLS_CONTROLLER}" -- ${cmd}
}

function is_rook_ceph_cluster_ha () {
  local osd_number
  osd_number=$(run_ceph_cmd "ceph status --format json"  | jq -r '.osdmap.num_osds')
  if [[ $osd_number -gt 2 ]]; then
    return 0
  else
    echo "Migration is not supported on non-HA ceph cluster. Current OSD count: '${osd_number}' Expected OSD count: '>2'"
    return 1
  fi
}

function is_replicated_data_pool_present() {
  if  run_ceph_cmd "ceph osd pool ls detail" | grep -q "'${ROOK_CEPH_DATA_POOL_NAME}' replicated "; then
    return 0
  else
    echo "Migration is not required as data pool: ${ROOK_CEPH_DATA_POOL_NAME} is not 'replicated' type"
    return 1
  fi
}

function is_rook_ceph_cluster_healthy() {
  local status
  status=$(run_ceph_cmd "ceph status --format json"  | jq -r '.health.status')
  if [[ $status == "HEALTH_OK" ]]; then
    return 0
  else
    echo "Rook ceph cluster is in unexpected state. Current state: '${status}' Expected state: 'HEALTH_OK'"
    return 1
  fi
}

function is_valid_number(){
  local number="$1"
  grep -q  -E '^[0-9]+$' <<< "${number}"
}

function is_ceph_storage_space_available(){
  local ceph_status
  local ceph_bytes_used
  local ceph_bytes_avail

  ceph_status=$(run_ceph_cmd "ceph status --format json")
  ceph_bytes_used=$(jq -r '.pgmap.bytes_used' <<< "${ceph_status}")
  ceph_bytes_avail=$(jq -r '.pgmap.bytes_avail' <<< "${ceph_status}")
  if ! is_valid_number "${ceph_bytes_used}"; then
    echo "Invalid Value for Ceph bytes used: '${ceph_bytes_used}'. Expected a valid number"
    return 1
  fi

  if ! is_valid_number "${ceph_bytes_avail}"; then
    echo "Invalid Value for Ceph bytes available: '${ceph_bytes_avail}'. Expected a valid number"
    return 1
  fi

  # Erasure coding uses 50% less storage space
  ERASURE_CODED_BYTES=$((ceph_bytes_used/2))
  if [[ $ceph_bytes_avail -gt $ERASURE_CODED_BYTES ]]; then
    echo "${ERASURE_CODED_BYTES}"
    return 0
  else
    echo "Ceph cluster does not enough storage capacity available to support migration. Required Capacity: $(numfmt --to=iec-i "${ERASURE_CODED_BYTES}") Available Capacity: $(numfmt --to=iec-i "${ceph_bytes_avail}"). Please increase OSD capacity by expanding OSD PVCs"
    return 1
  fi
}

function is_lh_storage_space_available() {
  local lh_nodes_data
  local lh_extra_capacity_per_osd_replica
  local osd_pv_ids_regex
  local osd_pv_lh_replica_dist
  local lh_node_name
  local lh_osd_replica_count
  local lh_node_storage_available
  local additional_storage_space_required
  local line

  # We use 3 OSDs
  lh_extra_capacity_per_osd_replica=$((ERASURE_CODED_BYTES/3))

  osd_pv_ids_regex=$(kubectl -n "${ROOK_CEPH_NAMESPACE}" get pvc -l ceph.rook.io/DeviceSet=set1 -o json | jq -r '[.items[].spec.volumeName] | join("|")')
  if [[ -z "${osd_pv_ids_regex}" ]]; then
    echo "Unable to find Ceph OSD PV IDs"
    return 1
  fi
  osd_pv_lh_replica_dist=$(kubectl -n $LONGHORN_NAMESPACE get replicas.longhorn.io  | grep -E "$osd_pv_ids_regex" | grep 'running' | awk '{print $3}' | sort | uniq -c)
  if [[ -z "${osd_pv_lh_replica_dist}" ]]; then
    echo "Unable to get meta-data of Running LH replicas for ceph OSD"
    return 1
  fi

  lh_nodes_data=$(kubectl -n "${LONGHORN_NAMESPACE}"  exec "$(kubectl -n "${LONGHORN_NAMESPACE}"  get po  -l "${LONGHORN_MANGER_LABELS}"  --field-selector status.phase=Running -o jsonpath='{.items[0].metadata.name}')" -- curl -s "http://longhorn-backend.${LONGHORN_NAMESPACE}:9500/v1/nodes")
  if [[ -z "${lh_nodes_data}" ]]; then
    echo "Unable to get longhorn nodes' meta-data"
    return 1
  fi

  while IFS=$'\n' read -r line ;do
    [[ -z "${line}" ]] && continue
    lh_node_name=$(echo "${line}" | awk '{print $2}')
    lh_osd_replica_count=$(echo "${line}" | awk '{print $1}')
    lh_node_storage_available=$(jq -r --arg nodeName "${lh_node_name}" --arg dataDiskPath "${LH_DATADISK_PATH}"  '.data[] | select(.name==$nodeName) | .disks | to_entries[] | select(.value.path==$dataDiskPath) | .value.storageAvailable' <<< "${lh_nodes_data}")
    if [[ -z "$lh_node_storage_available" ]]; then
      echo "Unable to get available storage for LH node: ${lh_node_name}"
      return 1
    fi

    if ! is_valid_number "${lh_node_storage_available}"; then
      echo "Invalid Value for Available storage for LH node: ${lh_node_name}"
      return 1
    fi
    additional_storage_space_required=$((lh_extra_capacity_per_osd_replica*lh_osd_replica_count))
    if [[ $lh_node_storage_available -lt $additional_storage_space_required ]]; then
      echo "Insufficient Storage space available for LH node: ${lh_node_name}. Required: $(numfmt --to=iec-i "${additional_storage_space_required}") Available: $(numfmt --to=iec-i "${lh_node_storage_available}")"
      return 2
    fi
  done <<< "${osd_pv_lh_replica_dist}"
}


function is_rook_ceph_tool_available() {
  if ! kubectl -n "${ROOK_CEPH_NAMESPACE}" get "${ROOK_CEPH_TOOLS_CONTROLLER}" >> /dev/null; then
    echo "Ceph tools deployment not found. Migration needs ceph tools deployment present in the cluster"
    return 1
  fi
  if ! kubectl -n "${ROOK_CEPH_NAMESPACE}" rollout status "${ROOK_CEPH_TOOLS_CONTROLLER}"; then
    echo "Ceph tools deployment not in healthy state. Migration needs healthy ceph tools deployment"
    return 1
  fi
}

function is_data_pool_present() {
  local pool_name="$1"
  if run_ceph_cmd "ceph osd pool ls --format json" | jq --arg poolName "${pool_name}" '. | index($poolName)' | grep -q 'null'; then
    echo "Ceph data pool: ${pool_name} does not exist in the cluster"
    return 1
  fi
}

function pre_migration_checks() {
  local return_msg

  if ! return_msg=$(is_rook_ceph_tool_available) ; then
    error "${return_msg}"
  fi

  if ! return_msg=$(is_rook_ceph_cluster_ha) ; then
    error "${return_msg}"
  fi

  if ! return_msg=$(is_rook_ceph_cluster_healthy) ; then
    error "${return_msg}"
  fi

  if ! return_msg=$(is_data_pool_present "${ROOK_CEPH_DATA_POOL_NAME}") ; then
    error "${return_msg}"
  fi

  if ! return_msg=$(is_replicated_data_pool_present) ; then
    info "${return_msg}"
    exit 0
  fi

  if ! return_msg=$(is_ceph_storage_space_available) ; then
    error "${return_msg}"
  fi
  ERASURE_CODED_BYTES="$return_msg"

  set +e
  return_msg=$(is_lh_storage_space_available)
  set -e
  case "$?" in
    0)
      echo ""
      ;;
    1)
      error "${return_msg}"
      ;;
    2)
      error_without_exit "${return_msg}"
      error_without_exit "Please increase storage space for LH i.e File System mounted on ${LH_DATADISK_PATH} to allow this script to perform ceph data pool migration"
      error "If it is not possible to increase the storage , please follow document to perform manual migration using Rados Export/Import"
      ;;
    *)
      error "Unexpected return code: $? for function: is_lh_storage_space_available"
      ;;
  esac
}

function is_ec_profile_available() {
  run_ceph_cmd "ceph osd erasure-code-profile get ${ROOK_CEPH_EC_PROFILE_NAME}" >> /dev/null 2>&1
}

function create_ec_profile() {
  run_ceph_cmd "ceph osd erasure-code-profile set ${ROOK_CEPH_EC_PROFILE_NAME} k=2 m=1  crush-failure-domain=host --force"
}

function create_ec_data_pool() {
  local pool_name="$1"
  run_ceph_cmd "ceph osd pool create ${pool_name} erasure ${ROOK_CEPH_EC_PROFILE_NAME}"
}

function get_crush_rule_name(){
  local pool_name="$1"
  local crush_rule_id
  local crush_rule_name

  crush_rule_id=$(run_ceph_cmd "ceph osd pool ls detail --format json" | jq -r --arg poolName "${pool_name}" '.[] | select(.pool_name==$poolName).crush_rule')
  [[ -z $crush_rule_id ]] && return
  crush_rule_name=$(run_ceph_cmd "ceph osd crush rule dump --format json"  | jq -r --arg ruleId "${crush_rule_id}" '.[] | select(.rule_id==($ruleId|tonumber)).rule_name')
  echo "${crush_rule_name}"
}

function delete_crush_rule(){
  local rule_name="$1"
  run_ceph_cmd "ceph osd crush rule rm ${rule_name}"
}

function rename_crush_rule() {
  local src_rule_name="$1"
  local dst_rule_name="$2"
  run_ceph_cmd "ceph osd crush rule rename ${src_rule_name} ${dst_rule_name}"
}

function delete_data_pool() {
  local pool_name="$1"
  local crush_rule_name

  crush_rule_name=$(get_crush_rule_name "${pool_name}")
  run_ceph_cmd "ceph osd pool rm ${pool_name} ${pool_name} --yes-i-really-really-mean-it --yes-i-really-really-mean-it-not-faking"
  [[ -z $crush_rule_name ]] && return
  delete_crush_rule "${crush_rule_name}"
}

function rename_data_pool() {
  local src_pool_name="$1"
  local dst_pool_name="$2"
  run_ceph_cmd "ceph osd pool rename ${src_pool_name} ${dst_pool_name}"
}

function get_pool_object_count(){
  local pool_name="$1"
  local pool_object_count

  pool_object_count=$(run_ceph_cmd "rados df --format json" | jq -r --arg poolName "${pool_name}" '.pools[] | select(.name==$poolName).num_objects')
  echo "${pool_object_count}"
}

function wait_until_pool_object_count_equal() {
  local src_pool_name="$1"
  local dst_pool_name="$2"
  local src_pool_object_count
  local dst_pool_object_count
  local try=120
  local index
  local return_code=1

  info "Verifying if Erasure coded pool and replicated pool has same object count"
  for index in $(seq 0 "${try}"); do
    src_pool_object_count=$(get_pool_object_count "${src_pool_name}")
    dst_pool_object_count=$(get_pool_object_count "${dst_pool_name}")
    if [[ $src_pool_object_count -eq $dst_pool_object_count ]]; then
      return_code=0
      break
    fi
    [[ $index -eq $try ]] || sleep 5
  done

  if [[ $return_code -eq 0 ]]; then
    echo "Found equal object count(${src_pool_object_count}) for pool '${src_pool_name}' '${dst_pool_name}'"
  else
    echo "Found difference in object count for pool '${src_pool_name}'(${src_pool_object_count}) '${dst_pool_name}'(${dst_pool_object_count})"
  fi
  return "${return_code}"
}

function scale_rook_ceph_operator() {
  local replica="${1:-0}"
  local argo_app_value="false"
  local scale_op="down"

  if [[ $replica -gt 0 ]]; then
    argo_app_value="true"
    scale_op="up"
  fi

  info "Setting selfHeal=${argo_app_value} on rook-ceph argo application"
  kubectl -n argocd patch application rook-ceph-operator --type=json -p '[{"op":"replace","path":"/spec/syncPolicy/automated/selfHeal","value":'${argo_app_value}'}]'
  kubectl -n argocd patch application rook-ceph-object-store --type=json -p '[{"op":"replace","path":"/spec/syncPolicy/automated/selfHeal","value":'${argo_app_value}'}]'
  kubectl -n argocd patch application fabric-installer --type=json -p '[{"op":"replace","path":"/spec/syncPolicy/automated/selfHeal","value":'${argo_app_value}'}]'
  info "Scaling ${scale_op} rook ceph operator"
  kubectl -n "${ROOK_CEPH_NAMESPACE}" scale --replicas="${replica}" deployment/rook-ceph-operator
  info "Waiting for rook ceph operator to scale ${scale_op}"

  if [[ $scale_op == "down" ]]; then
    while ! kubectl -n ${ROOK_CEPH_NAMESPACE} get pod -l app=rook-ceph-operator 2>&1  | grep -q 'No resources found'
    do
      sleep 5
      printf "."
    done
    echo ""
  else
    kubectl -n ${ROOK_CEPH_NAMESPACE} rollout status deployment/rook-ceph-operator
  fi
}

function enable_disable_ceph_traffic() {
  local action="$1"
  local check="${2:-false}"
  local policy_name="block-rook-ceph-rgw"
  local rgw_deploy
  local return_code
  case "$action" in
    enable)
      if [[ $check == "true" ]]; then
        if kubectl -n "${ROOK_CEPH_NAMESPACE}" get NetworkPolicy | grep  -q "^${policy_name} "; then
          return_code=1
        else
          return_code=0
        fi
        return "${return_code}"
      fi
      kubectl -n "${ROOK_CEPH_NAMESPACE}" delete NetworkPolicy "${policy_name}" --ignore-not-found
      ;;
    disable)
      if [[ $check == "true" ]]; then
        if kubectl -n "${ROOK_CEPH_NAMESPACE}" get NetworkPolicy | grep  -q "^${policy_name} "; then
          return_code=0
        else
          return_code=1
        fi
        return "${return_code}"
      fi
      kubectl apply -f - <<EOF
kind: NetworkPolicy
apiVersion: networking.k8s.io/v1
metadata:
  namespace: ${ROOK_CEPH_NAMESPACE}
  name: ${policy_name}
spec:
  podSelector:
    matchLabels:
  ingress:
  - from:
    - podSelector: {}
EOF
      ;;
    *)
      error "Invalid choice: ${action} for function: enable_disable_ceph_traffic"
      ;;
  esac
  for rgw_deploy in $(kubectl -n "${ROOK_CEPH_NAMESPACE}" get deploy -l app=rook-ceph-rgw  -o name);do
    kubectl -n "${ROOK_CEPH_NAMESPACE}" rollout restart "${rgw_deploy}"
    kubectl -n "${ROOK_CEPH_NAMESPACE}" rollout status "${rgw_deploy}"
  done
}

function migrate_data_pool(){
  local force
  local return_msg
  local epoch
  local temp_data_object_count
  local temp_epoch_object_count

  if ! is_ec_profile_available; then
    if [[ $GLOBAL_FORCE == "true" ]]; then
      force='yes'
    else
      warn "Erasure coded profile: '${ROOK_CEPH_EC_PROFILE_NAME}' does not exists on ceph cluster. The profile gets created automatically by the rook-ceph operator if data pool uses erasure coding"
      warn "If you wish to create the profile via migration script, type 'yes' and hit Enter for confirmation"
      read -r force </dev/tty
    fi
    if [[ $force == 'yes' ]] || [[ $force == "'yes'" ]]
    then
      warn "Creating Erasure coded profile: '${ROOK_CEPH_EC_PROFILE_NAME}'"
      if create_ec_profile; then
        info "Erasure coded profile: '${ROOK_CEPH_EC_PROFILE_NAME}' created successfully"
      else
        error "Erasure coded profile: '${ROOK_CEPH_EC_PROFILE_NAME}' creation failed"
      fi
    else
      error "Unexpected user input received: '${force}', aborting ceph data pool migration to erasure coding"
    fi
  fi

  if ! return_msg=$(is_data_pool_present "${EC_DATA_POOL_NAME}"); then
    info "Creating Erasure coded Pool: ${EC_DATA_POOL_NAME}"
    if ! create_ec_data_pool "${EC_DATA_POOL_NAME}"; then
      error "Erasure coded Pool: ${EC_DATA_POOL_NAME} creation failed"
    fi
  fi

  if ! run_ceph_cmd "ceph osd pool set ${EC_DATA_POOL_NAME} compression_mode none" ; then
    error "Failure while setting compression_mode=none for pool: '${EC_DATA_POOL_NAME}'"
  fi

  if ! run_ceph_cmd "ceph osd pool application enable ${EC_DATA_POOL_NAME} rook-ceph-rgw  --yes-i-really-mean-it" ; then
    error "Failure while setting application=rook-ceph-rgw for pool: '${EC_DATA_POOL_NAME}'"
  fi

  # Enable maintenance mode
  if ! enable_disable_ceph_traffic "disable" "true"; then
    if [[ $GLOBAL_FORCE == "true" ]]; then
      force='yes'
    else
      warn "Please be aware Ceph data pool migration will cause downtime, we are going to block object store during migration"
      warn "If you wish to continue, type 'yes' and hit Enter for confirmation."
      read -r force </dev/tty
    fi
    if [[ $force == 'yes' ]] || [[ $force == "'yes'" ]]; then
      ENABLE_CEPH_TRAFFIC="true"
      enable_disable_ceph_traffic "disable"
    else
      warn "Exiting migration as consent was not given to block object store"
      exit 1
    fi
  fi


  OBJECT_COUNT=$(get_pool_object_count "${ROOK_CEPH_DATA_POOL_NAME}")
  if ! is_valid_number "${OBJECT_COUNT}"; then
    error "Invalid value: '${OBJECT_COUNT}' found for object count in pool: '${ROOK_CEPH_DATA_POOL_NAME}'"
  fi

  info "Migrating Ceph data pool and logs are available at ${LOG_FILE}"
  DELETE_TEMP_EC_POOL="true"
  if run_ceph_cmd "rados cppool ${ROOK_CEPH_DATA_POOL_NAME} ${EC_DATA_POOL_NAME}" > "${LOG_FILE}" 2>&1; then
    if ! wait_until_pool_object_count_equal "${ROOK_CEPH_DATA_POOL_NAME}" "${EC_DATA_POOL_NAME}"; then
      error "Please re-run the script to re-initiate the migration to avoid difference in object count"
    fi
    if tail -n 1 "${LOG_FILE}" | grep -i -q "successfully copied pool ${ROOK_CEPH_DATA_POOL_NAME}"; then
      info "Successfully copied data from pool: ${ROOK_CEPH_DATA_POOL_NAME} to: ${EC_DATA_POOL_NAME}"
    else
      warn "Copy pool command executed successfully but could not find line 'Successfully copied data' in log file: ${LOG_FILE}"
    fi
  else
    error "Failure while running copy pool command"
  fi

  SCALE_UP_ROOK_CEPH_OPERATOR="true"
  scale_rook_ceph_operator "0"

  epoch=$(date "+%s")
  if rename_data_pool "${ROOK_CEPH_DATA_POOL_NAME}" "${ROOK_CEPH_DATA_POOL_NAME}.${epoch}"; then
    info "Successfully renamed pool from:${ROOK_CEPH_DATA_POOL_NAME} to:${ROOK_CEPH_DATA_POOL_NAME}.${epoch}"
  else
    warn "Phase 1 renaming operation failed. Reverting changes if any"
    temp_data_object_count=$(get_pool_object_count "${ROOK_CEPH_DATA_POOL_NAME}")
    if [[ $OBJECT_COUNT -eq $temp_data_object_count ]]; then
      info "[Revert] Deleting pool: '${ROOK_CEPH_DATA_POOL_NAME}.${epoch}' as pool: '${ROOK_CEPH_DATA_POOL_NAME}' contains original object count: '${OBJECT_COUNT}'"
      delete_data_pool "${ROOK_CEPH_DATA_POOL_NAME}.${epoch}"
    else
      temp_epoch_object_count=$(get_pool_object_count "${ROOK_CEPH_DATA_POOL_NAME}.${epoch}")
      if [[ $OBJECT_COUNT -eq $temp_epoch_object_count ]]; then
        info "[Revert] Deleting pool: '${ROOK_CEPH_DATA_POOL_NAME}' as pool: '${ROOK_CEPH_DATA_POOL_NAME}.${epoch}' contains original object count: '${OBJECT_COUNT}'"
        delete_data_pool "${ROOK_CEPH_DATA_POOL_NAME}"
        info "[Revert] Renaming pool from: '${ROOK_CEPH_DATA_POOL_NAME}.${epoch}' to: '${ROOK_CEPH_DATA_POOL_NAME}'"
        if ! rename_data_pool "${ROOK_CEPH_DATA_POOL_NAME}.${epoch}" "${ROOK_CEPH_DATA_POOL_NAME}"; then
          warn "[Revert] Unable to rename pool from: '${ROOK_CEPH_DATA_POOL_NAME}.${epoch}' to: '${ROOK_CEPH_DATA_POOL_NAME}'"
          warn "[Revert] Please do not run migration script again as it may cause data loss. Raise a support ticket with uipath to complete the migration"
          create_hard_lock
        fi
      else
        warn "[Revert] Unable to revert changes as none of the pool has original data pool object count: '${OBJECT_COUNT}'"
        warn "[Revert] Please do not run migration script again as it may cause data loss. Raise a support ticket with uipath to complete the migration"
        create_hard_lock
      fi
    fi
    error "Failure while renaming pool from:${ROOK_CEPH_DATA_POOL_NAME} to:${ROOK_CEPH_DATA_POOL_NAME}.${epoch}"
  fi

  if return_msg=$(is_data_pool_present "${ROOK_CEPH_DATA_POOL_NAME}"); then
    warn "Ceph data pool: ${ROOK_CEPH_DATA_POOL_NAME} should not exist but it does"
    temp_data_object_count=$(get_pool_object_count "${ROOK_CEPH_DATA_POOL_NAME}")
    if [[ -z $temp_data_object_count || $temp_data_object_count -eq 0 || $GLOBAL_FORCE == "true" ]]; then
      force='yes'
    else
      warn "Please verify if the pool: ${ROOK_CEPH_DATA_POOL_NAME} no longer contains important data. Type 'yes' if you wish to delete the pool"
      warn "Run command: 'kubect -n ${ROOK_CEPH_NAMESPACE} exec ${ROOK_CEPH_TOOLS_CONTROLLER} -- rados df' to view pool status"
      read -r force </dev/tty
    fi
    if [[ $force == 'yes' ]] || [[ $force == "'yes'" ]]
    then
      if ! delete_data_pool "${ROOK_CEPH_DATA_POOL_NAME}"; then
        error "Failure while deleting pool: ${ROOK_CEPH_TOOLS_CONTROLLER}"
      fi
    else
      warn "Exiting migration as pool: ${ROOK_CEPH_DATA_POOL_NAME} got recreated and consent was not given to delete the pool"
      exit 1
    fi
  fi

  if rename_data_pool "${EC_DATA_POOL_NAME}" "${ROOK_CEPH_DATA_POOL_NAME}"; then
    DELETE_TEMP_EC_POOL="false"
    info "Successfully renamed pool from:${EC_DATA_POOL_NAME} to:${ROOK_CEPH_DATA_POOL_NAME}"
  else
    warn "Phase 2 renaming operation failed. Reverting changes if any"
    info "[Revert] Deleting pool: '${ROOK_CEPH_DATA_POOL_NAME}' as we will be reverting previous rename action"
    delete_data_pool "${ROOK_CEPH_DATA_POOL_NAME}"
    if ! rename_data_pool "${ROOK_CEPH_DATA_POOL_NAME}.${epoch}" "${ROOK_CEPH_DATA_POOL_NAME}"; then
      warn "[Revert] Unable to revert previous pool rename from: '${ROOK_CEPH_DATA_POOL_NAME}.${epoch}' to:'${ROOK_CEPH_DATA_POOL_NAME}'.\nPlease do not run migration script again as it may cause data loss. Raise a support ticket with uipath to complete the migration"
      create_hard_lock
    fi
    error "Failure while renaming pool from:${EC_DATA_POOL_NAME} to:${ROOK_CEPH_DATA_POOL_NAME}"
  fi

  if ! delete_data_pool "${ROOK_CEPH_DATA_POOL_NAME}.${epoch}"; then
    info "Successfully deleted ceph pool: '${ROOK_CEPH_DATA_POOL_NAME}.${epoch}' after migration"
  fi

  crush_rule_name=$(get_crush_rule_name "${ROOK_CEPH_DATA_POOL_NAME}")
  if [[ -z "${crush_rule_name}" ]]; then
    warn "Unable to find crush rule name for pool: '${ROOK_CEPH_DATA_POOL_NAME}'"
  else
    delete_crush_rule "${ROOK_CEPH_DATA_POOL_NAME}"
    if ! rename_crush_rule "${crush_rule_name}" "${ROOK_CEPH_DATA_POOL_NAME}"; then
      warn "Unable to rename crush rule from: '${crush_rule_name}' to: '${ROOK_CEPH_DATA_POOL_NAME}'"
    else
      info "Successfully renamed crush rule from: '${crush_rule_name}' to: '${ROOK_CEPH_DATA_POOL_NAME}'"
    fi
  fi
}

function main() {
  parse_long_args "$@"
  check_hard_lock
  pre_migration_checks
  SKIP_CHECK_ON_EXIT="false"
  FAILURE_ON_EXIT="true"
  migrate_data_pool
  FAILURE_ON_EXIT="false"
}

main "$@"